from typing import Any, Hashable

from kumo.types import IndexSpec
from kumo.utils import get_nested


def get_index_value(data: Any, index_spec: IndexSpec) -> Hashable:
    if isinstance(index_spec, str):
        return get_nested(data, index_spec)
    return tuple(get_nested(data, s) for s in index_spec)


import heapq
import random
from typing import Any, Hashable, Sequence

from kumo.errors import DuplicatedError, EmptyStorageError
from kumo.storage.funcs import get_index_value
from kumo.types import (IndexSpec, Node, NodeData, NodeGetStrategy, NodeMeta,
                        PutBackStatus)


class MemoryCollection:
    def __init__(self, node: Node, unique_constraints: Sequence[IndexSpec], get_strategy: NodeGetStrategy):
        self.node = node
        self.data: list[tuple[NodeData, NodeMeta]] = []
        self.unique_indexes: dict[IndexSpec, dict[Hashable, int]] = {s: {} for s in unique_constraints}
        self.get_strategy = get_strategy
        self.new_set: set[int] = set()
        self.heap: list[tuple[float,float,int]] = []

    def put(self, data: NodeData):
        indexes_data = [(v, get_index_value(data, s)) for s, v in self.unique_indexes.items()]

        for index, index_value in indexes_data:
            if index_value in index:
                raise DuplicatedError()

        next_loc = len(self.data)
        
        for index, index_value in indexes_data:
            index[index_value] = next_loc
        
        self.new_set.add(next_loc)
        
        self.data.append((data, NodeMeta()))
    
    def get(self) -> NodeData:
        if len(self.new_set) > 0 and len(self.heap) > 0:
            if random.random() < self.get_strategy.exploration_prob:
                return self.explore()
            
            return self.exploit()
            
        if len(self.heap) > 0:
            return self.exploit()
        
        if len(self.new_set) > 0:
            return self.explore()
        
        raise EmptyStorageError(self.node)

    def explore(self) -> NodeData:
        loc = random.sample(self.new_set, 1)[0]
        self.new_set.remove(loc)

        return self.data[loc][0]
    
    def exploit(self) -> NodeData:
        return self.data[heapq.heappop(self.heap)[2]][0]

    def put_back(self, data: NodeData, status: PutBackStatus):
        loc = self.get_loc(data)

        data, meta = self.data[loc]

        meta.n_errors += status.n_errors
        meta.n_duplicated += status.n_duplicated
        meta.n_generated += status.n_generated
        meta.n_attempted += 1

        if self.get_strategy.max_errors is not None and meta.n_errors > self.get_strategy.max_errors:
            return

        if self.get_strategy.max_duplicated is not None and meta.n_duplicated > self.get_strategy.max_duplicated:
            return

        if meta.n_attempted >= self.get_strategy.n_exploration_attempts:
            heapq.heappush(self.heap, (-meta.n_generated / meta.n_attempted, random.random(), loc))
            return
        
        self.new_set.add(loc)
        
    def get_loc(self, data: NodeData) -> int:
        index_spec, index = next(iter(self.unique_indexes.items()))

        index_value = get_index_value(data, index_spec)

        loc = index[index_value]

        if self.data[loc][0] != data:
            raise ValueError(f"Data changed, {data} != {self.data[loc][0]}")

        return loc

class MemoryStorage:
    def __init__(self, get_strategy: NodeGetStrategy | None = None):
        self.nodes: list[Node] = []
        self.store: dict[Node, MemoryCollection] = {}
        self.get_strategy = NodeGetStrategy() if get_strategy is None else get_strategy
    
    def add_node(self, node: Node, unique_constraints: Sequence[str | tuple[str, ...]]):
        self.nodes.append(node)
        self.store[node] = MemoryCollection(node, unique_constraints, self.get_strategy)

    def get(self, node: Node) -> Any:
        return self.store[node].get()

    def put(self, data: Any):
        self.store[type(data)].put(data)

    def put_back(self, data: Any, status: PutBackStatus):
        self.store[type(data)].put_back(data, status)


import asyncio
import random
from asyncio import Queue
from typing import TypeVar

from kumo.edge import CrawlerEdge
from kumo.errors import DuplicatedError, EdgeError
from kumo.types import Counts, Edge, Node, PutBackStatus, Storage

E = TypeVar("E", bound=Edge)
N = TypeVar("N", bound=Node)


class Crawler:
    def __init__(self, storage: Storage):
        self.edges: list[CrawlerEdge] = []
        self.storage = storage

        self._requested: Queue[Counts] = Queue()
        self._available: Counts = {}

    def edge(self):
        def decorator(func: E) -> E:
            self.edges.append(CrawlerEdge.from_func(func))
            return func
        return decorator

    async def request(self, counts:Counts):
        await self._requested.put(counts)

    async def _process_edge(self, edge: CrawlerEdge):
        input_data = self.storage.get(edge.input_node)

        try:
            output_data = await edge(input_data)
        except EdgeError as e:
            print(e)
            self.storage.put_back(input_data, PutBackStatus(1, 0, 0))
            return

        n_duplicated = 0
        n_generated = 0

        for data in output_data:
            try:
                self.storage.put(data)
                n_generated += 1
            except DuplicatedError:
                n_duplicated += 1

        self.storage.put_back(input_data, PutBackStatus(0, n_duplicated, n_generated))                

    def strategy(self) -> CrawlerEdge:
        return random.choice(self.edges)

    async def loop(self):
        while True:
            edge = self.strategy()
            await self._process_edge(edge)
            
    def run(self):
        asyncio.run(self.loop())



from dataclasses import dataclass

from kumo.types import Edge, Node


@dataclass
class CrawlerEdge:
    input_node: Node
    output_node: Node
    func: Edge

    @classmethod
    def from_func(cls, func: Edge) -> "CrawlerEdge":
        annotations = func.__annotations__

        output_node = annotations.pop("return")

        if len(annotations) != 1:
            raise ValueError("Edge functions must have exactly one input")
        
        input_node = list(annotations.values())[0]

        return cls(input_node, output_node, func)

    async def __call__(self, input: Node) -> "list[Node]":
        return await self.func(input)



class EdgeError(Exception):
    """Exception raised when a node process fails to run."""

class DuplicatedError(Exception):
    """Raised when a duplicated item is found."""
    pass

class EmptyStorageError(Exception):
    """Raised when a storage is empty."""
    pass


# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['kumo', 'kumo.storage']

package_data = \
{'': ['*']}

install_requires = \
['pydantic>=1.10.2,<2.0.0']

setup_kwargs = {
    'name': 'kumo-py',
    'version': '0.1.0',
    'description': 'A small experimental framework for exploring graph-structured data.',
    'long_description': 'Small experimental library for exploring graph-structured data.\n\nTODO: Add a description of the project.\n',
    'author': 'marcinplatek',
    'author_email': 'maruplat@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/marcinplatek/kumo',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



from dataclasses import dataclass
from typing import Any, Callable, Coroutine, Dict, List, Protocol, Tuple, Union


@dataclass
class NodeMeta:
    n_errors: int = 0
    n_duplicated: int = 0
    n_generated: int = 0
    n_attempted: int = 0

@dataclass
class NodeGetStrategy:
    max_errors: "int | None" = 3
    max_duplicated: "int | None" = None
    exploration_prob: float = 0.5
    n_exploration_attempts: int = 3

Node = type

NodeData = Any

Edge = Callable[[NodeData], Coroutine[Any, Any, List[NodeData]]]

Counts = Dict[Node, int]

IndexSpec = Union[str, Tuple[str, ...]]

@dataclass
class PutBackStatus:
    n_errors: int = 0
    n_duplicated: int = 0
    n_generated: int = 0

class Storage(Protocol):
    def get(self, node: Node) -> NodeData:
        pass

    def put(self, data: NodeData):
        pass
    
    def put_back(self, data: NodeData, status: PutBackStatus):
        pass
